from __future__ import annotations

import shutil
import sys

from pathlib import Path
from typing import TYPE_CHECKING

import pytest

from cleo.formatters.style import Style
from cleo.io.buffered_io import BufferedIO
from poetry.core.packages.package import Package
from poetry.factory import Factory
from poetry.installation.operations.install import Install
from poetry.puzzle.exceptions import SolverProblemError
from poetry.repositories.repository import Repository
from poetry.repositories.repository_pool import RepositoryPool
from poetry.utils.env import MockEnv

from poetry_plugin_bundle.bundlers.venv_bundler import VenvBundler


if TYPE_CHECKING:
    from poetry.config.config import Config
    from poetry.poetry import Poetry
    from poetry.utils.env import VirtualEnv
    from pytest_mock import MockerFixture


@pytest.fixture()
def io() -> BufferedIO:
    io = BufferedIO()

    io.output.formatter.set_style("success", Style("green", options=["dark"]))
    io.output.formatter.set_style("warning", Style("yellow", options=["dark"]))

    return io


@pytest.fixture()
def poetry(config: Config) -> Poetry:
    poetry = Factory().create_poetry(
        Path(__file__).parent.parent / "fixtures" / "simple_project"
    )
    poetry.set_config(config)

    pool = RepositoryPool()
    repository = Repository("repo")
    repository.add_package(Package("foo", "1.0.0"))
    pool.add_repository(repository)
    poetry.set_pool(pool)

    return poetry


def _create_venv_marker_file(tempdir: str | Path) -> Path:
    marker_file = Path(tempdir) / "existing-venv-marker.txt"
    marker_file.write_text("This file should get deleted as part of venv recreation.")
    return marker_file


def test_bundler_should_build_a_new_venv_with_existing_python(
    io: BufferedIO, tmpdir: str, poetry: Poetry, mocker: MockerFixture
) -> None:
    shutil.rmtree(tmpdir)
    mocker.patch("poetry.installation.executor.Executor._execute_operation")

    bundler = VenvBundler()
    bundler.set_path(Path(tmpdir))

    assert bundler.bundle(poetry, io)

    expected = f"""\
  • Bundling simple-project (1.2.3) into {tmpdir}
  • Bundling simple-project (1.2.3) into {tmpdir}: Creating a virtual environment using Poetry-determined Python
  • Bundling simple-project (1.2.3) into {tmpdir}: Installing dependencies
  • Bundling simple-project (1.2.3) into {tmpdir}: Installing simple-project (1.2.3)
  • Bundled simple-project (1.2.3) into {tmpdir}
"""
    assert expected == io.fetch_output()


def test_bundler_should_build_a_new_venv_with_given_executable(
    io: BufferedIO, tmpdir: str, poetry: Poetry, mocker: MockerFixture
) -> None:
    shutil.rmtree(tmpdir)
    mocker.patch("poetry.installation.executor.Executor._execute_operation")

    bundler = VenvBundler()
    bundler.set_path(Path(tmpdir))
    bundler.set_executable(sys.executable)

    assert bundler.bundle(poetry, io)

    expected = f"""\
  • Bundling simple-project (1.2.3) into {tmpdir}
  • Bundling simple-project (1.2.3) into {tmpdir}: Creating a virtual environment using Python {sys.executable}
  • Bundling simple-project (1.2.3) into {tmpdir}: Installing dependencies
  • Bundling simple-project (1.2.3) into {tmpdir}: Installing simple-project (1.2.3)
  • Bundled simple-project (1.2.3) into {tmpdir}
"""
    assert expected == io.fetch_output()


def test_bundler_should_build_a_new_venv_if_existing_venv_is_incompatible(
    io: BufferedIO, tmpdir: str, poetry: Poetry, mocker: MockerFixture
) -> None:
    mocker.patch("poetry.installation.executor.Executor._execute_operation")

    mock_env = MockEnv(path=Path(tmpdir), is_venv=True, version_info=(1, 2, 3))
    mocker.patch("poetry.utils.env.EnvManager.get", return_value=mock_env)

    bundler = VenvBundler()
    bundler.set_path(Path(tmpdir))

    marker_file = _create_venv_marker_file(tmpdir)

    assert marker_file.exists()
    assert bundler.bundle(poetry, io)
    assert not marker_file.exists()

    expected = f"""\
  • Bundling simple-project (1.2.3) into {tmpdir}
  • Bundling simple-project (1.2.3) into {tmpdir}: Creating a virtual environment using Poetry-determined Python
  • Bundling simple-project (1.2.3) into {tmpdir}: Replacing existing virtual environment due to incompatible Python version
  • Bundling simple-project (1.2.3) into {tmpdir}: Installing dependencies
  • Bundling simple-project (1.2.3) into {tmpdir}: Installing simple-project (1.2.3)
  • Bundled simple-project (1.2.3) into {tmpdir}
"""
    assert expected == io.fetch_output()


def test_bundler_should_use_an_existing_venv_if_compatible(
    io: BufferedIO, tmp_venv: VirtualEnv, poetry: Poetry, mocker: MockerFixture
) -> None:
    mocker.patch("poetry.installation.executor.Executor._execute_operation")

    bundler = VenvBundler()
    bundler.set_path(tmp_venv.path)

    marker_file = _create_venv_marker_file(tmp_venv.path)

    assert marker_file.exists()
    assert bundler.bundle(poetry, io)
    assert marker_file.exists()

    path = str(tmp_venv.path)
    expected = f"""\
  • Bundling simple-project (1.2.3) into {path}
  • Bundling simple-project (1.2.3) into {path}: Creating a virtual environment using Poetry-determined Python
  • Bundling simple-project (1.2.3) into {path}: Installing dependencies
  • Bundling simple-project (1.2.3) into {path}: Installing simple-project (1.2.3)
  • Bundled simple-project (1.2.3) into {path}
"""
    assert expected == io.fetch_output()


def test_bundler_should_remove_an_existing_venv_if_forced(
    io: BufferedIO, tmp_venv: VirtualEnv, poetry: Poetry, mocker: MockerFixture
) -> None:
    mocker.patch("poetry.installation.executor.Executor._execute_operation")

    bundler = VenvBundler()
    bundler.set_path(tmp_venv.path)
    bundler.set_remove(True)

    marker_file = _create_venv_marker_file(tmp_venv.path)

    assert marker_file.exists()
    assert bundler.bundle(poetry, io)
    assert not marker_file.exists()

    path = str(tmp_venv.path)
    expected = f"""\
  • Bundling simple-project (1.2.3) into {path}
  • Bundling simple-project (1.2.3) into {path}: Creating a virtual environment using Poetry-determined Python
  • Bundling simple-project (1.2.3) into {path}: Installing dependencies
  • Bundling simple-project (1.2.3) into {path}: Installing simple-project (1.2.3)
  • Bundled simple-project (1.2.3) into {path}
"""
    assert expected == io.fetch_output()


def test_bundler_should_fail_when_installation_fails(
    io: BufferedIO, tmpdir: str, poetry: Poetry, mocker: MockerFixture
) -> None:
    mocker.patch(
        "poetry.installation.executor.Executor._do_execute_operation",
        side_effect=Exception(),
    )

    bundler = VenvBundler()
    bundler.set_path(Path(tmpdir))

    assert not bundler.bundle(poetry, io)

    expected = f"""\
  • Bundling simple-project (1.2.3) into {tmpdir}
  • Bundling simple-project (1.2.3) into {tmpdir}: Creating a virtual environment using Poetry-determined Python
  • Bundling simple-project (1.2.3) into {tmpdir}: Installing dependencies
  • Bundling simple-project (1.2.3) into {tmpdir}: Failed at step Installing dependencies
"""
    assert expected == io.fetch_output()


def test_bundler_should_display_a_warning_for_projects_with_no_module(
    io: BufferedIO, tmp_venv: VirtualEnv, mocker: MockerFixture, config: Config
) -> None:
    poetry = Factory().create_poetry(
        Path(__file__).parent.parent / "fixtures" / "simple_project_with_no_module"
    )
    poetry.set_config(config)

    pool = RepositoryPool()
    repository = Repository("repo")
    repository.add_package(Package("foo", "1.0.0"))
    pool.add_repository(repository)
    poetry.set_pool(pool)

    mocker.patch("poetry.installation.executor.Executor._execute_operation")

    bundler = VenvBundler()
    bundler.set_path(tmp_venv.path)
    bundler.set_remove(True)

    assert bundler.bundle(poetry, io)

    path = str(tmp_venv.path)
    expected = f"""\
  • Bundling simple-project (1.2.3) into {path}
  • Bundling simple-project (1.2.3) into {path}: Creating a virtual environment using Poetry-determined Python
  • Bundling simple-project (1.2.3) into {path}: Installing dependencies
  • Bundling simple-project (1.2.3) into {path}: Installing simple-project (1.2.3)
  • Bundled simple-project (1.2.3) into {path}
  • The root package was not installed because no matching module or package was found.
"""
    assert expected == io.fetch_output()


def test_bundler_can_filter_dependency_groups(
    io: BufferedIO, tmpdir: str, poetry: Poetry, mocker: MockerFixture, config: Config
) -> None:
    poetry = Factory().create_poetry(
        Path(__file__).parent.parent / "fixtures" / "simple_project_with_dev_dep"
    )
    poetry.set_config(config)

    # foo is in the main dependency group
    # bar is a dev dependency
    # add a repository for foo but not bar
    pool = RepositoryPool()
    repository = Repository("repo")
    repository.add_package(Package("foo", "1.0.0"))
    pool.add_repository(repository)
    poetry.set_pool(pool)

    mocker.patch("poetry.installation.executor.Executor._execute_operation")

    bundler = VenvBundler()
    bundler.set_path(Path(tmpdir))
    bundler.set_remove(True)

    # This should fail because there is not repo for bar
    with pytest.raises(SolverProblemError):
        assert not bundler.bundle(poetry, io)

    bundler.set_activated_groups({"main"})
    io.clear_output()

    # This succeeds because the dev dependency group is filtered out
    assert bundler.bundle(poetry, io)

    path = tmpdir
    expected = f"""\
  • Bundling simple-project (1.2.3) into {path}
  • Bundling simple-project (1.2.3) into {path}: Creating a virtual environment using Poetry-determined Python
  • Bundling simple-project (1.2.3) into {path}: Installing dependencies
  • Bundling simple-project (1.2.3) into {path}: Installing simple-project (1.2.3)
  • Bundled simple-project (1.2.3) into {path}
"""
    assert expected == io.fetch_output()


@pytest.mark.parametrize("compile", [True, False])
def test_bundler_passes_compile_flag(
    io: BufferedIO,
    tmp_venv: VirtualEnv,
    poetry: Poetry,
    mocker: MockerFixture,
    compile: bool,
) -> None:
    mocker.patch("poetry.installation.executor.Executor._execute_operation")

    bundler = VenvBundler()
    bundler.set_path(tmp_venv.path)
    bundler.set_remove(True)
    bundler.set_compile(compile)

    # bundle passes the flag from set_compile to enable_bytecode_compilation method
    mocker = mocker.patch(
        "poetry.installation.executor.Executor.enable_bytecode_compilation"
    )

    assert bundler.bundle(poetry, io)

    mocker.assert_called_once_with(compile)

    path = str(tmp_venv.path)
    expected = f"""\
  • Bundling simple-project (1.2.3) into {path}
  • Bundling simple-project (1.2.3) into {path}: Creating a virtual environment using Poetry-determined Python
  • Bundling simple-project (1.2.3) into {path}: Installing dependencies
  • Bundling simple-project (1.2.3) into {path}: Installing simple-project (1.2.3)
  • Bundled simple-project (1.2.3) into {path}
"""
    assert expected == io.fetch_output()


def test_bundler_editable_deps(
    io: BufferedIO, tmpdir: str, poetry: Poetry, mocker: MockerFixture, config: Config
) -> None:
    poetry = Factory().create_poetry(
        Path(__file__).parent.parent / "fixtures" / "simple_project_with_editable_dep"
    )
    poetry.set_config(config)

    install_spy = mocker.spy(Install, "__init__")
    mocker.patch("poetry.installation.executor.Executor._execute_operation")

    bundler = VenvBundler()
    bundler.set_path(Path(tmpdir))

    io.clear_output()

    bundler.bundle(poetry, io)

    path = tmpdir
    expected = f"""\
  • Bundling simple-project (1.2.3) into {path}
  • Bundling simple-project (1.2.3) into {path}: Creating a virtual environment using Poetry-determined Python
  • Bundling simple-project (1.2.3) into {path}: Installing dependencies
  • Bundling simple-project (1.2.3) into {path}: Installing simple-project (1.2.3)
  • Bundled simple-project (1.2.3) into {path}
"""
    assert expected == io.fetch_output()

    installed_packages = [call.args[1] for call in install_spy.call_args_list]
    dep_installs = list(
        filter(lambda package: package.name == "bar", installed_packages)
    )
    assert len(dep_installs) > 0

    editable_installs = list(filter(lambda package: package.develop, dep_installs))
    assert len(editable_installs) == 0
